@@ -2,7 +2,7 @@
 /* GdkPixbuf library - SVG image loader
  *
  * Copyright (C) 2002 Matthias Clasen
- * Copyright (C) 2002 Dom Lachowicz
+ * Copyright (C) 2002-2004 Dom Lachowicz
  *
  * Authors: Matthias Clasen <maclas@gmx.de>
  *          Dom Lachowicz <cinamod@hotmail.com>
@@ -25,14 +25,10 @@
 #include <stdlib.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gdk-pixbuf/gdk-pixbuf-io.h>
-
-#if HAVE_SVGZ
-#include <rsvg-gz.h>
-#endif
+#include "rsvg-private.h"
 
 typedef struct {
         RsvgHandle                 *handle;
-        GdkPixbuf                  *pixbuf;
 
         GdkPixbufModuleUpdatedFunc  updated_func;
         GdkPixbufModulePreparedFunc prepared_func;
@@ -46,9 +42,25 @@ typedef struct {
 G_MODULE_EXPORT void fill_vtable (GdkPixbufModule *module);
 G_MODULE_EXPORT void fill_info (GdkPixbufFormat *info);
 
+enum {
+        ERROR_WRITING = 1,
+        ERROR_DISPLAYING_IMAGE
+} RsvgLoaderErrorReasons;
+
+static void
+rsvg_propegate_error (GError ** err,
+                      const char * reason,
+                      gint code)
+{
+        if (err) {
+                *err = NULL;
+                g_set_error (err, rsvg_error_quark (), code, reason);
+        }
+}
+
 static gpointer
 gdk_pixbuf__svg_image_begin_load (GdkPixbufModuleSizeFunc size_func,
-                                  GdkPixbufModulePreparedFunc prepared_func, 
+                                  GdkPixbufModulePreparedFunc prepared_func,
                                   GdkPixbufModuleUpdatedFunc  updated_func,
                                   gpointer user_data,
                                   GError **error)
@@ -68,13 +80,30 @@ gdk_pixbuf__svg_image_begin_load (GdkPixbufModuleSizeFunc size_func,
         return context;
 }
 
+static void
+emit_updated (SvgContext *context, GdkPixbuf *pixbuf)
+{
+        if (context->updated_func != NULL)
+                (* context->updated_func) (pixbuf,
+                                           0, 0,
+                                           gdk_pixbuf_get_width (pixbuf),
+                                           gdk_pixbuf_get_height (pixbuf),
+                                           context->user_data);
+}
+
+static void
+emit_prepared (SvgContext *context, GdkPixbuf *pixbuf)
+{
+        if (context->prepared_func != NULL)
+                (* context->prepared_func) (pixbuf, NULL, context->user_data);
+}
+
 static gboolean
 gdk_pixbuf__svg_image_load_increment (gpointer data,
 				      const guchar *buf, guint size,
 				      GError **error)
 {
         SvgContext *context = (SvgContext *)data;
-        gboolean result;
 
         if (error)
                 *error = NULL;
@@ -82,54 +111,59 @@ gdk_pixbuf__svg_image_load_increment (gpointer data,
         if (context->first_write == TRUE) {
                 context->first_write = FALSE;
 
+                context->handle = rsvg_handle_new ();
 
-#ifdef HAVE_SVGZ
-                /* lazy create a SVGZ or SVG loader */
-                if ((size >= 2) && (buf[0] == (guchar)0x1f) && (buf[1] == (guchar)0x8b))
-                        context->handle = rsvg_handle_new_gz ();
-                else
-#endif
-                        context->handle = rsvg_handle_new ();
+                if (!context->handle) {
+                        rsvg_propegate_error (error, _("Error displaying image"), ERROR_DISPLAYING_IMAGE);
+                        return FALSE;
+                }
 
                 rsvg_handle_set_size_callback (context->handle, context->size_func, context->user_data, NULL);
         }
 
-        result = rsvg_handle_write (context->handle, buf, size, error);  
+        if (!context->handle) {
+                rsvg_propegate_error (error, _("Error displaying image"), ERROR_DISPLAYING_IMAGE);
+                return FALSE;
+        }
 
-        context->pixbuf = rsvg_handle_get_pixbuf (context->handle);
-  
-        if (context->pixbuf != NULL && context->prepared_func != NULL)
-                (* context->prepared_func) (context->pixbuf, NULL, context->user_data);        
-  
-        return result;
+        if (!rsvg_handle_write (context->handle, buf, size, error)) {
+                rsvg_propegate_error (error, _("Error writing"), ERROR_WRITING);
+                return FALSE;
+        }
+
+        return TRUE;
 }
 
 static gboolean
 gdk_pixbuf__svg_image_stop_load (gpointer data, GError **error)
 {
-        SvgContext *context = (SvgContext *)data;  
+        SvgContext *context = (SvgContext *)data;
+        GdkPixbuf *pixbuf;
+        gboolean result = TRUE;
 
         if (error)
                 *error = NULL;
 
-        rsvg_handle_close (context->handle, error);
-
-        if (context->pixbuf == NULL) {
-                context->pixbuf = rsvg_handle_get_pixbuf (context->handle);
-    
-                if (context->pixbuf != NULL && context->prepared_func != NULL)
-                        (* context->prepared_func) (context->pixbuf, NULL, context->user_data);
+        if (!context->handle) {
+                rsvg_propegate_error (error, _("Error displaying image"), ERROR_DISPLAYING_IMAGE);
+                return FALSE;
         }
 
-        if (context->pixbuf != NULL && context->updated_func != NULL)
-                (* context->updated_func) (context->pixbuf, 
-                                           0, 0, 
-                                           gdk_pixbuf_get_width (context->pixbuf), 
-                                           gdk_pixbuf_get_height (context->pixbuf), 
-                                           context->user_data);
+        rsvg_handle_close (context->handle, error);
+
+        pixbuf = rsvg_handle_get_pixbuf (context->handle);
+
+        if (pixbuf != NULL) {
+                emit_prepared (context, pixbuf);
+                emit_updated (context, pixbuf);
+                g_object_unref (pixbuf);
+        }
+        else {
+                rsvg_propegate_error (error, _("Error displaying image"), ERROR_DISPLAYING_IMAGE);
+                result = FALSE;
+        }
 
         rsvg_handle_free (context->handle);
-        g_object_unref (context->pixbuf);
         g_free (context);
 
         return TRUE;
@@ -138,40 +172,52 @@ gdk_pixbuf__svg_image_stop_load (gpointer data, GError **error)
 void
 fill_vtable (GdkPixbufModule *module)
 {
+        /*todo: call this somewhere to make sure the gnome-vfs et. al. are initialized */
+        /*rsvg_init ();*/
+
         module->begin_load     = gdk_pixbuf__svg_image_begin_load;
         module->stop_load      = gdk_pixbuf__svg_image_stop_load;
         module->load_increment = gdk_pixbuf__svg_image_load_increment;
 }
 
+/* this is present only in GTK+ 2.4 and later. we want librsvg to work with older versions too */
+#ifndef GDK_PIXBUF_FORMAT_SCALABLE
+#define GDK_PIXBUF_FORMAT_SCALABLE (1 << 1)
+#endif
+
+/* this is present only in GTK+ 2.6 and later. we want librsvg to work with older versions too. 
+   we can't define this flag yet because Pango isn't threadsafe. */
+#ifndef GDK_PIXBUF_FORMAT_THREADSAFE
+#define GDK_PIXBUF_FORMAT_THREADSAFE (1 << 2)
+#endif
+
 void
 fill_info (GdkPixbufFormat *info)
 {
         static GdkPixbufModulePattern signature[] = {
-                { "<?xml", NULL, 50 },
                 { "<svg", NULL, 100 },
                 { "<!DOCTYPE svg", NULL, 100 },
-#if HAVE_SVGZ
-                { "\x1f\x8b", NULL, 50 }, /* todo: recognizes any gzipped file, not much we can do */
-#endif
                 { NULL, NULL, 0 }
         };
-        static gchar *mime_types[] = { 
-                "image/svg", 
+        static gchar *mime_types[] = { /* yes folks, i actually have run into all of these in the wild... */
                 "image/svg+xml",
-                NULL 
+                "image/svg",
+                "image/svg-xml",
+                "image/vnd.adobe.svg+xml",
+                "text/xml-svg",
+                NULL
         };
-        static gchar *extensions[] = { 
-                "svg", 
-#if HAVE_SVGZ
+        static gchar *extensions[] = {
+                "svg",
                 "svgz",
-#endif
-                NULL 
+                "svg.gz",
+                NULL
         };
-        
+
         info->name        = "svg";
         info->signature   = signature;
-        info->description = "Scalable Vector Graphics";
+        info->description = _("Scalable Vector Graphics");
         info->mime_types  = mime_types;
         info->extensions  = extensions;
-        info->flags       = 0;
+        info->flags       = GDK_PIXBUF_FORMAT_SCALABLE;
 }
