@@ -180,148 +180,193 @@ print_pixbuf (GObject * ignored, gpointer user_data)
 
 #endif /* HAVE_GNOME_PRINT */
 
+#if GTK_CHECK_VERSION(2,4,0)
+
+static char *
+save_file (const char * title, GtkWidget * parent)
+{
+	GtkWidget *dialog;
+	char *filename = NULL;
+
+	dialog = gtk_file_chooser_dialog_new (title,
+										  GTK_WINDOW (parent),
+										  GTK_FILE_CHOOSER_ACTION_SAVE,
+										  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+										  GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
+										  NULL);
+
+	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
+		{
+			
+			filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
+		}
+	
+	gtk_widget_destroy (dialog);
+
+	return filename;
+}
+
+#else
+
+static char *
+save_file (const char * title, GtkWidget * parent)
+{
+	GtkWidget * filesel;
+	char * filename = NULL;
+			
+	filesel = gtk_file_selection_new (title);
+	gtk_window_set_transient_for(GTK_WINDOW(filesel), parent);
+	
+	if (gtk_dialog_run (GTK_DIALOG (filesel)) == GTK_RESPONSE_OK)
+		{			
+			filename = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel)));
+		}
+
+	gtk_widget_destroy (dialog);
+
+	return filename;
+}
+
+#endif
+
 static void 
 save_pixbuf (GObject * ignored, gpointer user_data)
 {
-	GtkWidget * filesel;
 	ViewerCbInfo * info = (ViewerCbInfo *)user_data;
+	char * filename;
+
+	filename = save_file (_("Save SVG as PNG"), info->window);
 	
-	filesel = gtk_file_selection_new (_("Save SVG as PNG"));
-	gtk_window_set_transient_for(GTK_WINDOW(filesel), GTK_WINDOW(info->window));
-	
-	if (gtk_dialog_run (GTK_DIALOG (filesel)) == GTK_RESPONSE_OK)
+	if (filename) 
 		{
-			const char * filename;
+			GError * err = NULL;
 			
-			filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel));
-			
-			if (filename) 
+			if (!gdk_pixbuf_save (info->pixbuf, filename, "png", &err, NULL)) 
 				{
-					GError * err = NULL;
-					
-					if (!gdk_pixbuf_save (info->pixbuf, filename, "png", &err, NULL)) 
+					if (err) 
 						{
-							if (err) 
-								{
-									GtkWidget * errmsg;
+							GtkWidget * errmsg;
+							
+							errmsg = gtk_message_dialog_new (GTK_WINDOW(info->window),
+															 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+															 GTK_MESSAGE_WARNING,
+															 GTK_BUTTONS_CLOSE,
+															 err->message);
 									
-									errmsg = gtk_message_dialog_new (GTK_WINDOW(filesel),
-																	 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-																	 GTK_MESSAGE_WARNING,
-																	 GTK_BUTTONS_CLOSE,
-																	 err->message);
-									gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
-									
-									gtk_dialog_run (GTK_DIALOG (errmsg));
-									
-									g_error_free (err);
-									gtk_widget_destroy (errmsg);
-								}
+							gtk_dialog_run (GTK_DIALOG (errmsg));
+							
+							g_error_free (err);
+							gtk_widget_destroy (errmsg);
 						}
-				} 
-			else 
-				{
-					GtkWidget * errmsg;
-					
-					errmsg = gtk_message_dialog_new (GTK_WINDOW(filesel),
-													 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-													 GTK_MESSAGE_WARNING,
-													 GTK_BUTTONS_CLOSE,
-													 _("No filename given"));
-					gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
-					
-					gtk_dialog_run (GTK_DIALOG (errmsg));
-					gtk_widget_destroy (errmsg);
 				}
-		}
-	
-	gtk_widget_destroy (filesel);
+
+			g_free (filename);
+		} 
+	else 
+		{
+			GtkWidget * errmsg;
+			
+			errmsg = gtk_message_dialog_new (GTK_WINDOW(info->window),
+											 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+											 GTK_MESSAGE_WARNING,
+											 GTK_BUTTONS_CLOSE,
+											 _("No filename given"));
+			gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
+					
+			gtk_dialog_run (GTK_DIALOG (errmsg));
+			gtk_widget_destroy (errmsg);
+		}	
 }
 
 static void 
 save_svg (GObject * ignored, gpointer user_data)
 {
-	GtkWidget * filesel;
 	ViewerCbInfo * info = (ViewerCbInfo *)user_data;
+	char * filename;
 	
-	filesel = gtk_file_selection_new (_("Save SVG"));
-	gtk_window_set_transient_for(GTK_WINDOW(filesel), GTK_WINDOW(info->window));
+	filename = save_file (_("Save SVG"), info->window);
 
-	if (gtk_dialog_run (GTK_DIALOG (filesel)) == GTK_RESPONSE_OK)
+	if (filename) 
 		{
-			const char * filename;
+			FILE * fp;
 			
-			filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel));
-			
-			if (filename) 
-				{
-					FILE * fp;
-
-					fp = fopen(filename, "wb");
-					if (!fp) 
-						{
-							GtkWidget * errmsg;
-							
-							errmsg = gtk_message_dialog_new (GTK_WINDOW(filesel),
-															 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-															 GTK_MESSAGE_WARNING,
-															 GTK_BUTTONS_CLOSE,
-															 _("Couldn't save %s"),
-															 filename);
-							gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
-									
-							gtk_dialog_run (GTK_DIALOG (errmsg));
-							gtk_widget_destroy (errmsg);
-						}
-					else
-						{
-							size_t written = 0, remaining = info->svg_bytes->len;
-							const char * buffer = info->svg_bytes->data;
-
-							while (remaining > 0) {
-								written = fwrite (buffer + (info->svg_bytes->len - remaining), 1, 
-												  remaining, fp);
-								if ((written < remaining) && ferror (fp) != 0)
-									{
-										GtkWidget * errmsg;
-										
-										errmsg = gtk_message_dialog_new (GTK_WINDOW(filesel),
-																		 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-																		 GTK_MESSAGE_WARNING,
-																		 GTK_BUTTONS_CLOSE,
-																		 _("Couldn't save %s"),
-																		 filename);
-										gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
-										
-										gtk_dialog_run (GTK_DIALOG (errmsg));
-										gtk_widget_destroy (errmsg);
-
-										break;
-									}
-								
-								remaining -= written;
-							}
-
-							fclose(fp);
-						}
-				} 
-			else 
+			/* todo: make this support gnome vfs */
+			fp = fopen(filename, "wb");
+			if (!fp) 
 				{
 					GtkWidget * errmsg;
 					
-					errmsg = gtk_message_dialog_new (GTK_WINDOW(filesel),
+					errmsg = gtk_message_dialog_new (GTK_WINDOW(info->window),
 													 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
 													 GTK_MESSAGE_WARNING,
 													 GTK_BUTTONS_CLOSE,
-													 _("No filename given"));
+													 _("Couldn't save %s"),
+													 filename);
 					gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
 					
 					gtk_dialog_run (GTK_DIALOG (errmsg));
 					gtk_widget_destroy (errmsg);
 				}
+			else
+				{
+					size_t written = 0, remaining = info->svg_bytes->len;
+					const char * buffer = info->svg_bytes->data;
+					
+					while (remaining > 0) {
+						written = fwrite (buffer + (info->svg_bytes->len - remaining), 1, 
+										  remaining, fp);
+						if ((written < remaining) && ferror (fp) != 0)
+							{
+								GtkWidget * errmsg;
+								
+								errmsg = gtk_message_dialog_new (GTK_WINDOW(info->window),
+																 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+																 GTK_MESSAGE_WARNING,
+																 GTK_BUTTONS_CLOSE,
+																 _("Couldn't save %s"),
+																 filename);
+								gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
+								
+								gtk_dialog_run (GTK_DIALOG (errmsg));
+								gtk_widget_destroy (errmsg);
+								
+								break;
+							}
+						
+						remaining -= written;
+					}
+
+					fclose(fp);
+				}
+
+			g_free (filename);
+		} 
+	else 
+		{
+			GtkWidget * errmsg;
+					
+			errmsg = gtk_message_dialog_new (GTK_WINDOW(info->window),
+											 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+											 GTK_MESSAGE_WARNING,
+											 GTK_BUTTONS_CLOSE,
+											 _("No filename given"));
+			gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));
+			
+			gtk_dialog_run (GTK_DIALOG (errmsg));
+			gtk_widget_destroy (errmsg);
 		}
-	
-	gtk_widget_destroy (filesel);
+}
+
+static void
+copy_svg_location(GObject * ignored, gpointer user_data)
+{
+	ViewerCbInfo * info = (ViewerCbInfo *)user_data;
+	GtkClipboard * clipboard = NULL;
+
+	if (info->base_uri) {
+		clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+		gtk_clipboard_set_text (clipboard, info->base_uri, -1);
+	}
 }
 
 static void
@@ -334,6 +379,19 @@ create_popup_menu (ViewerCbInfo * info)
 	popup_menu = gtk_menu_new ();
 	gtk_menu_set_accel_group (GTK_MENU (popup_menu), info->accel_group);
 
+	if (info->base_uri)
+	{
+		menu_item = gtk_image_menu_item_new_with_label (_("Copy SVG location"));
+		stock = gtk_image_new_from_stock (GTK_STOCK_COPY, GTK_ICON_SIZE_MENU);
+		gtk_widget_show (stock);
+		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), stock);
+		g_signal_connect (menu_item, "activate",
+						  G_CALLBACK (copy_svg_location), info);
+		gtk_widget_show (menu_item);
+		gtk_menu_shell_append (GTK_MENU_SHELL (popup_menu), menu_item);
+		gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_C, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
+	}
+
 	menu_item = gtk_image_menu_item_new_from_stock (GTK_STOCK_SAVE, NULL);
 	g_signal_connect (menu_item, "activate",
 					  G_CALLBACK (save_svg), info);
@@ -399,6 +457,7 @@ view_pixbuf (ViewerCbInfo * info, int xid, const char * color)
 	img_width = gdk_pixbuf_get_width (info->pixbuf);
 	img_height = gdk_pixbuf_get_height (info->pixbuf);
 
+#ifdef ENABLE_XEMBED
 	if(xid > 0)
 		{
 			GdkWindow *gdk_parent;
@@ -412,6 +471,7 @@ view_pixbuf (ViewerCbInfo * info, int xid, const char * color)
 			gtk_widget_add_events (win, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
 		}
 	else
+#endif
 		{
 			win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 
@@ -487,7 +547,6 @@ view_pixbuf (ViewerCbInfo * info, int xid, const char * color)
 int 
 main (int argc, char **argv)
 {
-	FILE * in;
 	GError * err = NULL;
 	poptContext popt_context;
 	double x_zoom = 1.0;
@@ -510,19 +569,19 @@ main (int argc, char **argv)
 	struct poptOption options_table[] = 
 		{
 #ifdef ENABLE_XEMBED
-			{ "xid",         'i',  POPT_ARG_INT,    &xid,         0, _("XWindow ID [for X11 embedding]"), _("<int>") },
+			{ "xid",         'i',  POPT_ARG_INT,    &xid,         0, N_("XWindow ID [for X11 embedding]"), N_("<int>") },
 #endif
-			{ "stdin",       's',  POPT_ARG_NONE,   &from_stdin,  0, _("Read from stdin instead of a file"), NULL },
-			{ "dpi-x",       'd',  POPT_ARG_DOUBLE, &dpi_x,       0, _("Set the # of Pixels Per Inch"), _("<float>") },
-			{ "dpi-y",       'p',  POPT_ARG_DOUBLE, &dpi_y,       0, _("Set the # of Pixels Per Inch"), _("<float>") },
-			{ "x-zoom",      'x',  POPT_ARG_DOUBLE, &x_zoom,      0, _("Set the x zoom factor"), _("<float>") },
-			{ "y-zoom",      'y',  POPT_ARG_DOUBLE, &y_zoom,      0, _("Set the y zoom factor"), _("<float>") },
-			{ "width",       'w',  POPT_ARG_INT,    &width,       0, _("Set the image's width"), _("<int>") },
-			{ "height",      'h',  POPT_ARG_INT,    &height,      0, _("Set the image's height"), _("<int>") },
-			{ "bg-color",    'b',  POPT_ARG_STRING, &bg_color,    0, _("Set the image background color (default: transparent)"), _("<string>") },
-			{ "base-uri",    'u',  POPT_ARG_STRING, &base_uri,    0, _("Set the base URI (default: none)"), _("<string>") },
-			{ "keep-aspect", 'k',  POPT_ARG_NONE,   &bKeepAspect, 0, _("Preserve the image's aspect ratio"), NULL },
-			{ "version",     'v',  POPT_ARG_NONE,   &bVersion,    0, _("Show version information"), NULL },
+			{ "stdin",       's',  POPT_ARG_NONE,   &from_stdin,  0, N_("Read from stdin instead of a file"), NULL },
+			{ "dpi-x",       'd',  POPT_ARG_DOUBLE, &dpi_x,       0, N_("Set the # of Pixels Per Inch"), N_("<float>") },
+			{ "dpi-y",       'p',  POPT_ARG_DOUBLE, &dpi_y,       0, N_("Set the # of Pixels Per Inch"), N_("<float>") },
+			{ "x-zoom",      'x',  POPT_ARG_DOUBLE, &x_zoom,      0, N_("Set the x zoom factor"), N_("<float>") },
+			{ "y-zoom",      'y',  POPT_ARG_DOUBLE, &y_zoom,      0, N_("Set the y zoom factor"), N_("<float>") },
+			{ "width",       'w',  POPT_ARG_INT,    &width,       0, N_("Set the image's width"), N_("<int>") },
+			{ "height",      'h',  POPT_ARG_INT,    &height,      0, N_("Set the image's height"), N_("<int>") },
+			{ "bg-color",    'b',  POPT_ARG_STRING, &bg_color,    0, N_("Set the image background color (default: transparent)"), N_("<string>") },
+			{ "base-uri",    'u',  POPT_ARG_STRING, &base_uri,    0, N_("Set the base URI (default: none)"), N_("<string>") },
+			{ "keep-aspect", 'k',  POPT_ARG_NONE,   &bKeepAspect, 0, N_("Preserve the image's aspect ratio"), NULL },
+			{ "version",     'v',  POPT_ARG_NONE,   &bVersion,    0, N_("Show version information"), NULL },
 			POPT_AUTOHELP
 			POPT_TABLEEND
 		};
@@ -534,7 +593,6 @@ main (int argc, char **argv)
 	info.svg_bytes = NULL;
 	info.window = NULL;
 	info.popup_menu = NULL;
-	info.base_uri = base_uri;
 
 	popt_context = poptGetContext ("rsvg-view", argc, (const char **)argv, options_table, 0);
 	poptSetOtherOptionHelp (popt_context, _("[OPTIONS...] [file.svg]"));
@@ -563,8 +621,9 @@ main (int argc, char **argv)
 	
 	/* initialize gtk+ */
 	gtk_init (&argc, &argv) ;
+	rsvg_init ();
 
-	rsvg_set_default_dpi_x_y (dpi_x, dpi_y);
+	rsvg_set_default_dpi (dpi_x, dpi_y);
 	
 	/* if both are unspecified, assume user wants to zoom the pixbuf in at least 1 dimension */
 	if (width == -1 && height == -1)
@@ -592,14 +651,42 @@ main (int argc, char **argv)
 
 	size_data.keep_aspect_ratio = bKeepAspect;
 	
-	if(from_stdin)
-		in = stdin;
-	else {
-		in = fopen (args[0], "rb");
-		base_uri = (char *)args[0];
+	if(!from_stdin) {
+		if (base_uri == NULL)
+			base_uri = (char *)args[0];
+	
+		info.svg_bytes = _rsvg_acquire_xlink_href_resource(args[0], base_uri, NULL);
+	} else {
+		info.svg_bytes = g_byte_array_new ();
+
+		for (;;)
+			{
+				char buf[1024 * 8];
+				size_t nread = fread (buf, 1, sizeof(buf), stdin);
+			
+				if (nread > 0)
+					g_byte_array_append (info.svg_bytes, buf, nread);
+				
+				if (nread < sizeof (buf))
+				{
+					if (ferror (stdin))
+						{
+							g_critical (_("Error reading\n"));
+							g_byte_array_free (info.svg_bytes, TRUE);
+							poptFreeContext (popt_context);
+							fclose(stdin);
+							
+							return 1;
+						}
+					else if (feof(stdin))
+						break;
+				}
+			}
+		
+		fclose(stdin);		
 	}
 
-	if(!in)
+	if(!info.svg_bytes || !info.svg_bytes->len)
 		{
 			g_critical (_("Couldn't open %s\n"), args[0]);
 			poptFreeContext (popt_context);
@@ -607,33 +694,7 @@ main (int argc, char **argv)
 			return 1;
 		}
 
-	info.svg_bytes = g_byte_array_new ();
-
-	for (;;)
-		{
-			char buf[1024 * 8];
-			size_t nread = fread (buf, 1, sizeof(buf), in);
-			
-			if (nread > 0)
-				g_byte_array_append (info.svg_bytes, buf, nread);
-			
-			if (nread < sizeof (buf))
-				{
-					if (ferror (in))
-						{
-							g_critical (_("Error reading\n"));
-							g_byte_array_free (info.svg_bytes, TRUE);
-							poptFreeContext (popt_context);
-							fclose(in);
-							
-							return 1;
-						}
-					else if (feof(in))
-						break;
-				}
-		}
-	
-	fclose(in);
+	info.base_uri = base_uri;
 
 	poptFreeContext (popt_context);
 
@@ -661,6 +722,7 @@ main (int argc, char **argv)
 
 	g_object_unref (G_OBJECT (info.pixbuf));
 	g_byte_array_free (info.svg_bytes, TRUE);
+	rsvg_term ();
 
 	return 0;
 }
